import re
from collections import deque

from pandoc.types import Str, Space, SoftBreak, Strong, Emph, Strikeout, Code, Link, \
    Underline, Math, InlineMath

from n2y.property_values import simplify_date


class PlainText():
    def __init__(self, text):
        self.text = text

    def to_pandoc(self):
        """Tokenize the text"""
        ast = []
        match = re.findall(r"( +)|(\S+)|(\n+)|(\t+)", self.text)

        for m in match:
            space, word, newline, tab = m
            for _ in range(len(space)):
                ast.append(Space())
            if word:
                ast.append(Str(word))
            for _ in range(len(newline)):
                ast.append(SoftBreak())
            for _ in range(len(tab) * 4):  # 4 spaces per tab
                ast.append(Space())
        return ast


class Annotations():
    def __init__(self, block):
        for key, value in block.items():
            self.__dict__[key] = value

    def listify(self, obj):
        return obj if type(obj) == list else [obj]

    def apply_pandoc(self, target):
        prependages = deque()
        appendages = deque()
        problematic_annotations = [
            self.bold, self.italic, self.strikethrough
        ]

        if True in problematic_annotations:
            blank_space = [Space(), SoftBreak()]
            while target[0] in blank_space:
                prependages.append(target.pop(0))
            while target[-1] in blank_space:
                appendages.appendleft(target.pop(-1))

        result = target
        if self.code:
            result = self.listify(Code(("", [], []), result))
        if self.bold:
            result = self.listify(Strong(result))
        if self.italic:
            result = self.listify(Emph(result))
        if self.underline:
            result = self.listify(Underline(result))
        if self.strikethrough:
            result = self.listify(Strikeout(result))

        return list(prependages) + result + list(appendages)


class InlineEquation():
    def __init__(self, block):
        for key, value in block.items():
            self.__dict__[key] = value

    def to_pandoc(self):
        return [Math(InlineMath(), self.expression)]


class Mention():
    """
    Just display the name of the mentions; one can imagine various other ways
    one may want to display these; e.g., using links.
    """

    def __init__(self, block):
        for key, value in block.items():
            self.__dict__[key] = value

    def to_pandoc(self):
        if self.type == 'user':
            return [Str(self.user["name"])]
        if self.type == 'page':
            # TODO: at least replace with the page title
            # TODO: incorporate when implementing https://github.com/innolitics/n2y/issues/24
            page_id = self.page['id']
            return [Str(f'Link to page "{page_id}"')]
        if self.type == 'database':
            # TODO: at least replace with the database title
            # TODO: incorporate when implementing https://github.com/innolitics/n2y/issues/24
            database_id = self.database['id']
            return [Str(f'Link to database "{database_id}"')]
        if self.type == 'date':
            return [Str(simplify_date(self.date))]
        if self.type == 'link_preview':
            return [Link(
                ('', [], []),
                self.plain_text.to_pandoc(),
                (self.href, ''))]
        else:
            raise NotImplementedError(f'Unknown mention type: "{self.type}"')


class RichText():
    def __init__(self, block):
        handlers = {
            'annotations': Annotations,
            'plain_text': PlainText,
            'equation': InlineEquation,
            'mention': Mention,
        }
        # TODO: Replace this loop with explicit attribute setters
        for key, value in block.items():
            if key in handlers.keys():
                self.__dict__[key] = handlers[key](value)
            else:
                self.__dict__[key] = value

    def to_pandoc(self):
        if self.type == 'text':
            if self.annotations.code:
                # Send raw text if it's code.
                return self.annotations.apply_pandoc(self.plain_text.text)
            elif 'href' in self.__dict__ and self.href:
                # links
                return [Link(
                    ('', [], []),
                    self.annotations.apply_pandoc(self.plain_text.to_pandoc()),
                    (self.href, ''))]
            else:
                # regular text
                return self.annotations.apply_pandoc(self.plain_text.to_pandoc())
        elif self.type == 'equation':
            return self.equation.to_pandoc()
        elif self.type == 'mention':
            return self.mention.to_pandoc()
        else:
            raise NotImplementedError(f'Unknown rich text object type: "{self.type}"')


class RichTextArray:
    def __init__(self, text):
        self.items = [RichText(i) for i in text]

    def to_pandoc(self):
        return sum([item.to_pandoc() for item in self.items], [])



from itertools import groupby
import logging
import importlib.util
from os import path

from pandoc.types import (
    Str, Para, Plain, Header, CodeBlock, BulletList, OrderedList, Decimal,
    Period, Meta, Pandoc, Link, HorizontalRule, BlockQuote, Image, MetaString,
    Table, TableHead, TableBody, TableFoot, RowHeadColumns, Row, Cell, RowSpan,
    ColSpan, ColWidthDefault, AlignDefault, Caption, Math, DisplayMath,
)

from n2y.notion import Client
from n2y.rich_text import RichText, RichTextArray


# Notes:
# A single Notion block may have multiple lines of text.
# A page is a block that puts children into its "content" attribute.
# We transform page blocks into other block types.
#
# Pandoc makes each word a block, and spaces are blocks too!
#
# Block types used here that do not exist in Notion:
#   container - block with no top-level content, only children (used to parse a page and lists)
#   bulleted_list - Notion has bulleted_list_item, but no enclosing container
#   numbered_list - Notion has numbered_list_item, but no enclusing container


logger = logging.getLogger(__name__)


def load_plugins(filename):
    # TODO: Consider storing the imported classes on the `Client` instance. This
    # will make it easier to test the plugin system in isolation, since given
    # that the plugins are globals right now, if we have a test that uses the
    # plugins it will mutate the global module state. Furthermore, by moving the
    # plugin block class references onto the `Client` class it will make it
    # easier to dynamically enable or disable certain plugin classes for certain
    # notion pages, which is a feature we will likely need in the future.

    # TODO: Make it possible to swap out the RichText classes too

    # TODO: Make it possible to modify the property value code too; note that we
    # should probably create a hierarchy of classes, similar to how the Block
    # classes work

    abs_path = path.abspath(filename)
    plugin_spec = importlib.util.spec_from_file_location("plugins", abs_path)
    plugin_module = importlib.util.module_from_spec(plugin_spec)
    plugin_spec.loader.exec_module(plugin_module)
    for (key, value) in plugin_module.exports.items():
        if key in globals():
            class_to_replace = globals()[key]
            plugin_base_class_names = [b.__name__ for b in value.__bases__]
            # plugins can only override classes in this file that are derrived from a Block
            if class_to_replace.__name__ in plugin_base_class_names \
                    and issubclass(class_to_replace, Block):
                globals()[key] = value
            else:
                logger.warning(
                    'Cannot import plugin "%s" because it is not '
                    'derrived from a known class.', key)
        else:
            raise NotImplementedError(f'Unknown plugin type "{key}".')


def load_block(client: Client, id, get_children=True):
    block = client.get_block(id)
    return parse_block(client, block, get_children)


# The Notion client object is passed down for the following reasons:
#   1. Some child objects may be unknown until the block is processed.
#      Links to other Notion pages are an example.
#   2. In some cases a block may choose not to get child blocks.
#      Currently, all blocks load all children.
def parse_block(client: Client, block, get_children=True):
    if block['type'] == "child_page":
        return ChildPageBlock(client, block, get_children)
    elif block['type'] == "paragraph":
        return ParagraphBlock(client, block, get_children)
    elif block['type'] == "heading_1":
        return HeadingOneBlock(client, block, get_children)
    elif block['type'] == "heading_2":
        return HeadingTwoBlock(client, block, get_children)
    elif block['type'] == "heading_3":
        return HeadingThreeBlock(client, block, get_children)
    elif block['type'] == "divider":
        return DividerBlock(client, block, get_children)
    elif block['type'] == "numbered_list_item":
        return NumberedListItemBlock(client, block, get_children)
    elif block['type'] == "bulleted_list_item":
        return BulletedListItemBlock(client, block, get_children)
    elif block['type'] == "to_do":
        return ToDoListItemBlock(client, block, get_children)
    elif block['type'] == "bookmark":
        return BookmarkBlock(client, block, get_children)
    elif block['type'] == "image":
        return ImageBlock(client, block, get_children)
    elif block['type'] == "code":
        return FencedCodeBlock(client, block, get_children)
    elif block['type'] == "quote":
        return QuoteBlock(client, block, get_children)
    elif block['type'] == "table":
        return TableBlock(client, block, get_children)
    elif block['type'] == "table_row":
        return RowBlock(client, block, get_children)
    elif block['type'] == "toggle":
        return ToggleBlock(client, block, get_children)
    elif block['type'] == "equation":
        return EquationBlock(client, block, get_children)
    elif block['type'] == "callout":
        return CalloutBlock(client, block, get_children)
    else:
        # TODO: add remaining block types
        raise NotImplementedError(f'Unknown block type: "{block["type"]}"')


class Block:
    def __init__(self, client: Client, block, get_children=True):
        logger.debug("Instantiating %s block", type(self).__name__)
        self.client = client

        self.notion_id = block['id']
        self.created_time = block['created_time']
        self.created_by = block['created_by']
        self.last_edited_time = block['last_edited_time']
        self.last_edited_by = block['last_edited_by']
        self.has_children = block['has_children']
        self.archived = block['archived']
        self.notion_type = block['type']
        self.notion_data = block[block['type']]

        if self.has_children and get_children:
            self.get_children()

    def to_pandoc(self):
        raise NotImplementedError()

    def children_to_pandoc(self):
        assert self.has_children
        pandoc_ast = []
        for block_type, blocks in groupby(self.children, lambda c: type(c)):
            if issubclass(block_type, ListItemBlock):
                pandoc_ast.append(block_type.list_to_pandoc(blocks))
            else:
                pandoc_ast.extend(b.to_pandoc() for b in blocks)
        return pandoc_ast

    def get_children(self):
        self.children = []
        for child in self.client.get_block_children(self.notion_id, recursive=False):
            self.children.append(parse_block(self.client, child, get_children=True))


class ListItemBlock(Block):
    @classmethod
    def list_to_pandoc(klass, items):
        raise NotImplementedError()


class ChildPageBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.title = self.notion_data["title"]

    def to_pandoc(self):
        if hasattr(self, 'children'):
            children = self.children_to_pandoc()
            return Pandoc(Meta({'title': MetaString(self.title)}), children)
        else:
            return None


class EquationBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.expression = self.notion_data["expression"]

    def to_pandoc(self):
        return Para([Math(DisplayMath(), self.expression)])


class ParagraphBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.text = RichTextArray(self.notion_data["text"])

    def to_pandoc(self):
        content = self.text.to_pandoc()
        if self.has_children:
            result = [Para(content)]
            children = self.children_to_pandoc()
            result.extend(children)
        else:
            result = Para(content)
        return result


class BulletedListItemBlock(ListItemBlock):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.text = RichTextArray(self.notion_data["text"])

    def to_pandoc(self):
        content = [Plain(self.text.to_pandoc())]
        if self.has_children:
            children = self.children_to_pandoc()
            content.extend(children)
        return content

    @classmethod
    def list_to_pandoc(klass, blocks):
        return BulletList([b.to_pandoc() for b in blocks])


class ToDoListItemBlock(BulletedListItemBlock):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.checked = self.notion_data['checked']
        if self.checked:
            self.text.items[0].plain_text.text = '☒ ' + self.text.items[0].plain_text.text
        else:
            self.text.items[0].plain_text.text = '☐ ' + self.text.items[0].plain_text.text


class NumberedListItemBlock(ListItemBlock):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.text = RichTextArray(self.notion_data['text'])

    def to_pandoc(self):
        content = [Plain(self.text.to_pandoc())]
        if self.has_children:
            children = self.children_to_pandoc()
            content.extend(children)
        return content

    @classmethod
    def list_to_pandoc(klass, blocks):
        return OrderedList((1, Decimal(), Period()), [b.to_pandoc() for b in blocks])


class HeadingBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.text = RichTextArray(self.notion_data["text"])

    def to_pandoc(self):
        return Header(self.level, ('', [], []), self.text.to_pandoc())


class HeadingOneBlock(HeadingBlock):
    level = 1


class HeadingTwoBlock(HeadingBlock):
    level = 2


class HeadingThreeBlock(HeadingBlock):
    level = 3


class DividerBlock(Block):
    def to_pandoc(self):
        return HorizontalRule()


class BookmarkBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.url = self.notion_data["url"]
        if self.notion_data["caption"]:
            self.caption = RichTextArray(self.notion_data["caption"])
        else:
            self.caption = None

    def to_pandoc(self):
        if self.caption:
            caption_ast = self.caption.to_pandoc()
        else:
            caption_ast = [Str(self.url)]
        return Para([Link(('', [], []), caption_ast, (self.url, ''))])


class FencedCodeBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.language = self.notion_data["language"]
        self.text = RichText(self.notion_data["text"][0])

    def to_pandoc(self):
        return CodeBlock(('', [self.language], []), self.text.plain_text.text)


class QuoteBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.text = RichTextArray(self.notion_data["text"])

    def to_pandoc(self):
        return BlockQuote([Para(self.text.to_pandoc())])


class ImageBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.file = File(client, block['image'])
        self.caption = RichTextArray(self.notion_data["caption"])

    def to_pandoc(self):
        url = None
        if self.file.type == "external":
            url = self.file.url
        elif self.file.type == "file":
            url = self.file.download()
        return Para([Image(('', [], []), self.caption.to_pandoc(), (url, ''))])


class TableBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.has_column_header = self.notion_data['has_column_header']
        self.has_row_header = self.notion_data['has_row_header']
        self.table_width = self.notion_data['table_width']

    def to_pandoc(self):
        children = self.children_to_pandoc()
        # Isolate the header row if it exists
        if self.has_column_header:
            header_rows = [children.pop(0)]
        else:
            header_rows = []
        if self.has_row_header:
            row_header_columns = 1
        else:
            row_header_columns = 0
        # Notion does not have cell alignment or width options, sticking with defaults.
        colspec = [(AlignDefault(), ColWidthDefault()) for _ in range(self.table_width)]
        table = Table(
            ('', [], []),
            Caption(None, []),
            colspec,
            TableHead(('', [], []), header_rows),
            [TableBody(
                ('', [], []),
                RowHeadColumns(row_header_columns),
                [],
                children
            )],
            TableFoot(('', [], []), [])
        )
        return table


class RowBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.cells = [RichTextArray(cell) for cell in self.notion_data["cells"]]

    def to_pandoc(self):
        cells = [Cell(
            ('', [], []),
            AlignDefault(),
            RowSpan(1),
            ColSpan(1),
            [Plain(cell.to_pandoc())]
        ) for cell in self.cells]
        return Row(('', [], []), cells)


class ToggleBlock(Block):
    """
    Generates a bulleted list item with indented children. A plugin may be used
    to add html classes and replicate the interactive behavior found in Notion.
    """

    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.text = RichTextArray(self.notion_data["text"])

    def to_pandoc(self):
        header = self.text.to_pandoc()
        children = self.children_to_pandoc()
        content = [Para(header)]
        content.extend(children)
        return BulletList([content])


class CalloutBlock(Block):
    def __init__(self, client: Client, block, get_children=True):
        super().__init__(client, block, get_children)
        self.text = RichTextArray(self.notion_data["text"])

    def to_pandoc(self):
        content = self.text.to_pandoc()
        if self.has_children:
            children = self.children_to_pandoc()
            result = [Para(content)]
            result.extend(children)
        else:
            result = Para(content)
        return result


class File:
    """
    See https://developers.notion.com/reference/file-object
    """

    def __init__(self, client: Client, obj):
        self.client = client
        if obj['type'] == "file":
            self.type = "file"
            self.url = obj['file']['url']
            self.expiry_time = obj['file']['expiry_time']
        elif obj['type'] == "external":
            self.type = "external"
            self.url = obj['external']['url']

    def download(self):
        return self.client.download_file(self.url)



from enum import Enum


class RequestTimeoutError(Exception):
    """Exception for requests that timeout.
    The request that we made waits for a specified period of time or maximum number of
    retries to get the response. But if no response comes within the limited time or
    retries, then this Exception is raised.
    """

    code = "request_timeout"

    def __init__(self, message: str = "Request to Notion API has timed out") -> None:
        super().__init__(message)


class HTTPResponseError(Exception):
    """Exception for HTTP errors.
    Responses from the API use HTTP response codes that are used to indicate general
    classes of success and error.
    """

    def __init__(self, response, message=None) -> None:
        if message is None:
            message = (
                f"Request to Notion API failed with status: {response.status_code}"
            )
        super().__init__(message)
        self.status = response.status_code
        self.headers = response.headers
        self.body = response.text


class APIErrorCode(str, Enum):
    Unauthorized = "unauthorized"
    """The bearer token is not valid."""

    RestrictedResource = "restricted_resource"
    """Given the bearer token used, the client doesn't have permission to
    perform this operation."""

    ObjectNotFound = "object_not_found"
    """Given the bearer token used, the resource does not exist.
    This error can also indicate that the resource has not been shared with owner
    of the bearer token."""

    RateLimited = "rate_limited"
    """This request exceeds the number of requests allowed. Slow down and try again."""

    InvalidJSON = "invalid_json"
    """The request body could not be decoded as JSON."""

    InvalidRequestURL = "invalid_request_url"
    """The request URL is not valid."""

    InvalidRequest = "invalid_request"
    """This request is not supported."""

    ValidationError = "validation_error"
    """The request body does not match the schema for the expected parameters."""

    ConflictError = "conflict_error"
    """The transaction could not be completed, potentially due to a data collision.
    Make sure the parameters are up to date and try again."""

    InternalServerError = "internal_server_error"
    """An unexpected error occurred. Reach out to Notion support."""

    ServiceUnavailable = "service_unavailable"
    """Notion is unavailable. Try again later.
    This can occur when the time to respond to a request takes longer than 60 seconds,
    the maximum request timeout."""


class APIResponseError(HTTPResponseError):
    """An error raised by Notion API."""

    def __init__(self, response, message, code) -> None:
        super().__init__(response, message)
        self.code = code


def is_api_error_code(code: str) -> bool:
    """Check if given code belongs to the list of valid API error codes."""
    if isinstance(code, str):
        return code in (error_code.value for error_code in APIErrorCode)
    return False


# Some of this code was taken from https://github.com/ramnes/notion-sdk-py
'''
The MIT License (MIT)

Copyright (c) 2021 Guillaume Gelin

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''



"""
Grabbing data from the Notion API
"""
import logging
import json
from os import path, makedirs
from shutil import copyfileobj
from urllib.parse import urlparse, urljoin

import requests

from n2y.errors import HTTPResponseError, APIResponseError, is_api_error_code, APIErrorCode


logger = logging.getLogger(__name__)


class Client:
    def __init__(self, access_token, media_root='.', media_url=''):
        self.access_token = access_token
        self.media_root = media_root
        self.media_url = media_url
        self.base_url = "https://api.notion.com/v1/"
        self.headers = {
            "Authorization": f"Bearer {self.access_token}",
            "Content-Type": "application/json",
            "Notion-Version": "2021-08-16",
        }

    def get_page_or_database(self, object_id):
        """
        First attempt to get the page corresponding with the object id; if
        the page doesn't exist, then attempt to retrieve the database. This
        trial-and-error is necessary because the API doesn't provide a means to
        determining what type of object corresponds with an ID and we don't want
        to make the user indicate if they are pulling down a database or a page.
        """
        try:
            return self.get_page(object_id), "page"
        except APIResponseError as e:
            if e.code == APIErrorCode.ObjectNotFound:
                return self.get_database(object_id), "database"
            else:
                raise e

    def get_database(self, database_id):
        starting_url = f"{self.base_url}databases/{database_id}/query"

        def depaginator(url):
            while True:
                data = self._post_url(url)
                yield data["results"]
                if not data["has_more"]:
                    return
                else:
                    url = data["next_cursor"]

        return sum(depaginator(starting_url), [])

    def get_page(self, page_id):
        return self._get_url(f"{self.base_url}pages/{page_id}")

    def get_block(self, block_id):
        url = f"{self.base_url}blocks/{block_id}"
        response = requests.get(url, headers=self.headers)
        return self._parse_response(response)

    def get_block_children(self, block_id, recursive=False):
        """Recursively get block children"""
        starting_url = f"{self.base_url}blocks/{block_id}/children"

        # Blocks that may have children.
        # https://developers.notion.com/reference/block
        blocks_to_expand = [
            "paragraph",
            "callout",
            "quote",
            "bulleted_list_item",
            "numbered_list_item",
            "to_do",
            "toggle",
            "column_list",
            "column",
            "template",
            "synced_block",
            "table",
        ]

        def depaginator(url):
            while True:
                data = self._get_url(url)
                yield data["results"]
                if not data["has_more"]:
                    return
                else:
                    cursor = data["next_cursor"]
                    url = f"{starting_url}?start_cursor={cursor}"

        result = sum(depaginator(starting_url), [])

        if recursive:
            for item in result:
                if item['has_children'] and item['type'] in blocks_to_expand:
                    item['children'] = self.get_block_children(item['id'])

        return result

    def _get_url(self, url):
        response = requests.get(url, headers=self.headers)
        return self._parse_response(response)

    def _post_url(self, url):
        response = requests.post(url, headers=self.headers)
        return self._parse_response(response)

    def _parse_response(self, response):
        """Taken from https://github.com/ramnes/notion-sdk-py"""
        try:
            response.raise_for_status()
        except requests.HTTPError as error:
            try:
                body = error.response.json()
                code = body.get("code")
            except json.JSONDecodeError:
                code = None
            if code and is_api_error_code(code):
                raise APIResponseError(response, body["message"], code)
            raise HTTPResponseError(error.response)
        body = response.json()
        logger.debug("=> %s", body)
        return body

    def download_file(self, url):
        # TODO: append created time as hex to end of file to prevent collisions?
        makedirs(self.media_root, exist_ok=True)
        url_path = path.basename(urlparse(url).path)
        local_filename = path.join(self.media_root, url_path)
        with requests.get(url, stream=True) as request_stream:
            with open(local_filename, 'wb') as file_stream:
                copyfileobj(request_stream.raw, file_stream)
        return urljoin(self.media_url, url_path)


def id_from_share_link(share_link):
    hyphens_removed = share_link.replace("-", "")
    if not hyphens_removed.startswith("https://www.notion.so/"):
        return hyphens_removed
    else:
        return hyphens_removed.split("/")[-1].split("?")[0]



import os
import re
import sys
import logging
import argparse

import yaml
import pandoc

from n2y import blocks, notion, property_values

logger = None


def cli_main():
    args = sys.argv[1:]
    access_token = os.environ.get('NOTION_ACCESS_TOKEN', None)
    sys.exit(main(args, access_token))


def main(raw_args, access_token):
    parser = argparse.ArgumentParser(
        description="Move data from Notion into YAML/markdown",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("object_id", help="The id or url for a Notion database or page")
    parser.add_argument(
        "--format", '-f',
        choices=["yaml", "markdown"], default="yaml",
        help=(
            "Select output type (only applies to databases)\n"
            "  yaml - log yaml to stdout\n"
            "  markdown - create a markdown file for each page"))
    parser.add_argument(
        "--media-root", help="Filesystem path to directory where images and media are saved")
    parser.add_argument("--media-url", help="URL for media root; must end in slash if non-empty")
    parser.add_argument("--plugins", '-p', help="Plugin file")
    parser.add_argument(
        "--output", '-o', default='./',
        help="Relative path to output directory")
    parser.add_argument(
        "--verbosity", '-v', default='INFO',
        help="Level to set the root logging module to")
    parser.add_argument(
        "--logging-format", default='%(asctime)s - %(levelname)s: %(message)s',
        help="Default format used when logging")
    parser.add_argument(
        "--name-column", '-n', default='title',
        help=(
            "Database column that will be used to generate the filename "
            "for each row. Column names are normalized to lowercase letters, "
            "numbers, and underscores. Only used when generating markdown."))
    args = parser.parse_args(raw_args)

    logging.basicConfig(
        format=args.logging_format, level=logging.__dict__[args.verbosity])
    global logger
    logger = logging.getLogger(__name__)

    if access_token is None:
        logger.critical('No NOTION_ACCESS_TOKEN environment variable is set')
        return 1

    object_id = notion.id_from_share_link(args.object_id)
    media_root = args.media_root or args.output
    if args.plugins:
        blocks.load_plugins(args.plugins)

    client = notion.Client(access_token, media_root, args.media_url)

    object_data, object_type = client.get_page_or_database(object_id)

    # TODO: in the future, determing the natural keys for each row in the
    # database and calculate them up-front; prune out any pages where the
    # natural key is empty. Furthermore, add duplicate handling here. Once the
    # natural key handling is done, there should be no need for the
    # `name_column_valid` since that will be handled here

    if object_type == "database" and args.format == 'markdown':
        if name_column_valid(object_data, args.name_column):
            export_database_as_markdown_files(client, object_data, options=args)
        else:
            return 1
    elif object_type == "database" and args.format == 'yaml':
        export_database_as_yaml_file(client, object_data)
    elif object_type == "page":
        export_page_as_markdown(client, object_data)

    return 0


def name_column_valid(raw_rows, name_column):
    first_row_flattened = property_values.flatten_property_values(raw_rows[0])

    def available_columns():
        return filter(
            lambda c: isinstance(first_row_flattened[c], str),
            first_row_flattened.keys())

    # make sure the title column exists
    if name_column not in first_row_flattened:
        logger.critical(
            'Database does not contain the column "%s". Please specify '
            'the correct name column using the --name-column (-n) flag. '
            # only show columns that have strings as possible options
            'Available column(s): ' + ', '.join(available_columns()), name_column)
        return False

    # make sure title column is not empty (only the first row is checked)
    if first_row_flattened[name_column] is None:
        logger.critical('Column "%s" Cannot Be Empty.', name_column)
        return False

    # make sure the title column is a string
    if not isinstance(first_row_flattened[name_column], str):
        logger.critical(
            'Column "%s" does not contain a string. '
            # only show columns that have strings as possible options
            'Available column(s): ' + ', '.join(available_columns()), name_column)
        return False
    return True


def export_database_as_markdown_files(client, raw_rows, options):
    os.makedirs(options.output, exist_ok=True)
    file_names = []
    skips = {'unnamed': 0, 'duplicate': 0}
    for row in raw_rows:
        meta = property_values.flatten_property_values(row)
        page_name = meta[options.name_column]
        if page_name:
            # sanitize file name just a bit
            # maybe use python-slugify in the future?
            filename = re.sub(r"[\s/,\\]", '_', page_name.lower())
            if filename not in file_names:
                file_names.append(filename)

                pandoc_output = blocks.load_block(client, row['id']).to_pandoc()
                logger.info('Processing page "%s".', page_name)

                markdown = pandoc_tree_to_markdown(pandoc_output)

                with open(os.path.join(options.output, f"{filename}.md"), 'w') as f:
                    write_yaml_frontmatter(f, meta)
                    f.write(markdown)
            else:
                logger.debug('Page name "%s" has been used', page_name)
                skips['duplicate'] += 1
        else:
            skips['unnamed'] += 1
    for key, count in skips.items():
        if count > 0:
            logger.info("%d %s page(s) skipped", count, key)


def write_yaml_frontmatter(open_file, metadata):
    open_file.write('---\n')
    # TODO: replace with nice clean YAML dumper
    open_file.write(yaml.dump(metadata))
    open_file.write('---\n\n')


def export_database_as_yaml_file(client, raw_rows):
    result = []
    for row in raw_rows:
        pandoc_output = blocks.load_block(client, row['id']).to_pandoc()
        markdown = pandoc_tree_to_markdown(pandoc_output) if pandoc_output else None
        result.append({**property_values.flatten_property_values(row), 'content': markdown})

    print(yaml.dump(result, sort_keys=False))


def export_page_as_markdown(client, page):
    pandoc_output = blocks.load_block(client, page['id']).to_pandoc()
    markdown = pandoc_tree_to_markdown(pandoc_output) if pandoc_output else None
    metadata = property_values.flatten_property_values(page)

    # For now, print the single page to standard output; eventually we'll need
    # to re-think this and likely write the result to a file. This is necessary
    # if there are sub-pages or sub-databases which would need to go into
    # separate files.
    write_yaml_frontmatter(sys.stdout, metadata)
    print(markdown)


def pandoc_tree_to_markdown(pandoc_tree):
    return pandoc.write(pandoc_tree, format='gfm+tex_math_dollars') \
        .replace('\r\n', '\n')  # Deal with Windows line endings


if __name__ == "__main__":
    cli_main()



"""
Simplify data that has been returned from the Notion API.
"""
import logging

logger = logging.getLogger(__name__)


def flatten_database_rows(raw_rows):
    return [flatten_property_values(r) for r in raw_rows]


def flatten_property_values(raw_data):
    return {
        k: simplify_property(v)
        for k, v in raw_data["properties"].items()
    }


def simplify_property(prop):
    if prop["type"] == "title":
        return simplify_title(prop["title"])
    elif prop["type"] == "rich_text":
        return simplify_rich_text(prop["rich_text"])
    elif prop["type"] == "select":
        return simplify_select(prop["select"])
    elif prop["type"] == "people":
        return simplify_people(prop["people"])
    elif prop["type"] == "url":
        return prop["url"]
    elif prop["type"] == "number":
        return prop["number"]
    elif prop["type"] == "email":
        return prop["email"]
    elif prop["type"] == "checkbox":
        return prop["checkbox"]
    elif prop["type"] == "phone_number":
        return prop["phone_number"]
    elif prop["type"] == "date":
        return simplify_date(prop["date"])
    elif prop["type"] == "multi_select":
        return [option["name"] for option in prop["multi_select"]]
    elif prop["type"] == "relation":
        logger.warning("field type 'relation' is not fully implemented")
        return [relation["id"] for relation in prop["relation"]]
    elif prop["type"] == "rollup":
        logger.warning("field type 'rollup' is not fully implemented")
        r = prop["rollup"]
        return {"type": r["type"], "function": r["function"], r["type"]: r[r["type"]]}
    elif prop["type"] == "formula":
        logger.warning("field type 'formula' is not fully implemented")
        return prop["formula"]["string"]
    elif prop["type"] == "files":
        logger.warning("field type 'files' is not fully implemented")
        return prop["files"]
    else:
        # TODO: add remaining column types
        raise NotImplementedError()


def simplify_title(data):
    if data is None or len(data) == 0:
        return None
    return data[0]["plain_text"]


def simplify_select(data):
    if data is None:
        return None
    return data["name"]


def simplify_rich_text(data):
    # Note that this will strip out some things that markdown doesn't support,
    # like colors and underlines.
    return "".join(simplify_rich_text_item(r) for r in data)


def simplify_rich_text_item(data):
    text = escape_markdown(data["plain_text"])
    if data["annotations"]["code"]:
        text = f"`{text}`"
    if data["annotations"]["bold"]:
        text = f"**{text}**"
    if data["annotations"]["italic"]:
        text = f"*{text}*"
    if data["annotations"]["strikethrough"]:
        text = f"~~{text}~~"
    if data["href"] is not None:
        text = f"[{text}]({data['href']})"
    return text


def escape_markdown(text):
    # TODO: think through other things we should escape and if/when we need to
    # escape
    escaped = "*_`[]"
    return "".join(c if c not in escaped else f"\\{c}" for c in text)


def simplify_people(data):
    return [p["name"] for p in data]


def simplify_date(data):
    if data is None:
        return None
    if data["end"] is None:
        return data["start"]
    else:
        return f'{data["start"]} to {data["end"]}'
